// *****************************************************************************
// *                                                                           *
// *  BiqCrunch is a semidefinite-based solver for binary quadratic problems.  *
// *  It uses a branch-and-bound method featuring an improved semidefinite     *
// *  bounding procedure, mixed with a polyhedral approach. BiqCrunch uses     *
// *  particular input files format to describe the combinatorial problems.    *
// *                                                                           *
// *   Copyright (C) 2010-2014 Nathan Krislock, Jérôme Malick, Frédéric Roupin *
// *                                                                           *
// *                 http://www-lipn.univ-paris13.fr/BiqCrunch/                *
// *									       *
// *****************************************************************************
//									       *
//    This program is free software: you can redistribute it and/or modify     *
//    it under the terms of the GNU General Public License as published by     *
//    the Free Software Foundation, either version 3 of the License, or        *
//    (at your option) any later version.                                      *
//                                                                             *
//    This program is distributed in the hope that it will be useful,          *
//    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
//    GNU General Public License for more details.                             *
//                                                                             *
//    You should have received a copy of the GNU General Public License        *
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
//                                                                             *
// *****************************************************************************
// Authors : N. Krislock, F. Roupin 2010, updated in 2012

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <memory.h>
#include <math.h>

#include "bb.h"
#include "biqcrunch.h"

/**
 * Name of the heuristic printed in the output file
 */
char heur_name[] = "Generic";


/**
 * Allocate the user defined global structure data (called at the beginning of the algorithm)
 * @param P0: original problem in domain [-1:1]
 * @param N: number of variables
 * @return 1 for success 0 for failure
 */
int BC_allocHeuristic(Problem *P0) {
	return 1;
}
/**
 * Free the user defined global structure data (called at the end of the algorithm)
 */
void BC_freeHeuristic() {
}

double primalHeuristic(Problem *P0, int *x);
double sdpBoundHeuristic(Problem *P0, BobNode * n, int * x);
double roundingHeuristic(Problem *P0, BobNode * n, int * x);

/**
 * Generic call of the heuristics.
 * This function is called at different times:
 * - at the beginning of the algorithm (PRIMAL_HEUR)
 * - during the evaluation of a node bound (SDP_BOUND_HEUR)
 * - after the evaluation of a node (ROUNDING_HEUR)
 * It has to return the value obtained during the execution of the chosen heuristic and the binary vector of the solution
 *
 * @param P0: the original problem in domain [-1:1]
 * @param node: the current node () (NULL if heuristic_code equals to PRIMAL_HEUR)
 * @param N: the total number of variables
 * @param x: output binary vector of the solution obtained from the heuristic in domain [0:1]
 * @param heuristic_code: code that identify the timing of the call (PRIMAL_HEUR, SDP_BOUND_HEUR, ROUNDING_HEUR)
 * @param max_problem: 1 if we're solving a maximization problem, 0 if we're solving a minimization problem
 * @return value of the solution found with the heuristic
 */
double BC_runHeuristic(Problem *P0, Problem *P, BobNode *node, int *x, int heuristic_code) {
	double heur_val;

	switch (heuristic_code) {

        case PRIMAL_HEUR:

            heur_val = primalHeuristic(P0, x);
            break;

        case SDP_BOUND_HEUR:

            heur_val = sdpBoundHeuristic(P0, node, x);
            break;

        case ROUNDING_HEUR:
            
            heur_val = roundingHeuristic(P0, node, x);
            break;

        default:
            printf("Choosen heuristic doesn't exist\n");
            exit(1);
	}

	return heur_val;
}

double primalHeuristic(Problem *P0, int *x) {
	int N = P0->n - 1;
	int i;
	double random;
    double gamma;
    double heur_val;
	int temp_x[N]; // temporary vector to store heuristic solutions
    double best;

	srand(time(NULL));

	if (P0->max_problem) best = -BIG_NUMBER; else best = BIG_NUMBER;

    for (gamma = 0.0; gamma < 1.0; gamma += 0.01) {

        for (i = 0; i < N; i++) {
            random = ((double) rand() / (double) RAND_MAX);

            if (random > gamma) {
                temp_x[i] = 1;
            } else {
                temp_x[i] = 0;
            }
        }    

        heur_val = BC_evaluateSolution(temp_x);

        if (((P0->max_problem && best < heur_val)
            || (!P0->max_problem && best > heur_val))
            && BC_isFeasibleSolution(temp_x)) {
            
            memcpy(x, temp_x, sizeof(int) * N);
            best = heur_val;
        }
    }

	return best;
}

double sdpBoundHeuristic(Problem *P0, BobNode * node, int * x) {
    int N = P0->n - 1;
	int i;
    int count;
    double heur_val;
	int temp_x[N]; // temporary vector to store heuristic solutions
    double best;
    double cut_off;

	srand(time(NULL));
    
	if (P0->max_problem) best = -BIG_NUMBER; else best = BIG_NUMBER;

    // Using a deterministic cut-off
    for (count=0; count<N; count++) {
        cut_off = node->fracsol[count];

        for (i = 0; i < N; i++) {
            if (node->xfixed[i]) {
                temp_x[i] = node->sol.X[i];
            } else {
                if (node->fracsol[i] > cut_off) {
                    temp_x[i] = 1;
                } else {
                    temp_x[i] = 0;
                }
            }
        }

        heur_val = BC_evaluateSolution(temp_x);

		if (((P0->max_problem && best < heur_val)
				|| (!P0->max_problem && best > heur_val))
				&& BC_isFeasibleSolution(temp_x)) {
			memcpy(x, temp_x, sizeof(int) * N);
			best = heur_val;
		}
    }

    // Using a different random cut-off for each xi
    for (count=0; count<100; count++) {

        for (i = 0; i < N; i++) {
            cut_off = ((double) rand() / (double) RAND_MAX); // var cut-off
            if (node->xfixed[i]) {
                temp_x[i] = node->sol.X[i];
            } else {
                if (node->fracsol[i] > cut_off) {
                    temp_x[i] = 1;
                } else {
                    temp_x[i] = 0;
                }
            }
        }

        heur_val = BC_evaluateSolution(temp_x);

		if (((P0->max_problem && best < heur_val)
				|| (!P0->max_problem && best > heur_val))
				&& BC_isFeasibleSolution(temp_x)) {
			memcpy(x, temp_x, sizeof(int) * N);
			best = heur_val;
		}
    }

	return best;
}


double roundingHeuristic(P0, node, x)
	Problem *P0; // Instance of the problem
	BobNode * node; // Node of the tree
	int *x;
{
	int N = P0->n - 1;
	int i;
    int count;
	int temp_x[N]; // temporary vector to store heuristic solutions
    double best;
    double cut_off;
    double heur_val;

	srand(time(NULL));

	if (P0->max_problem) best = -BIG_NUMBER; else best = BIG_NUMBER;

    // Using a deterministic cut-off
    for (count=0; count<N; count++) {
        cut_off = node->fracsol[count];

        for (i = 0; i < N; i++) {
            if (node->xfixed[i]) {
                temp_x[i] = node->sol.X[i];
            } else {
                if (node->fracsol[i] > cut_off) {
                    temp_x[i] = 1;
                } else {
                    temp_x[i] = 0;
                }
            }
        }

        heur_val = BC_evaluateSolution(temp_x);

        if (((P0->max_problem && best < heur_val)
                || (!P0->max_problem && best > heur_val))
                && BC_isFeasibleSolution(temp_x)) {
            memcpy(x, temp_x, sizeof(int) * N);
            best = heur_val;
        }
    }

    // Using a different random cut-off for each xi
    for (count=0; count<100; count++) {

        for (i = 0; i < N; i++) {
            cut_off = ((double) rand() / (double) RAND_MAX); // var cut_off
            if (node->xfixed[i]) {
                temp_x[i] = node->sol.X[i];
            } else {
                if (node->fracsol[i] > cut_off) {
                    temp_x[i] = 1;
                } else {
                    temp_x[i] = 0;
                }
            }
        }

        heur_val = BC_evaluateSolution(temp_x);

        if (((P0->max_problem && best < heur_val)
                || (!P0->max_problem && best > heur_val))
                && BC_isFeasibleSolution(temp_x)) {
            memcpy(x, temp_x, sizeof(int) * N);
            best = heur_val;
        }
    }
	return best;
}
