// *****************************************************************************
// *                                                                           *
// *  BiqCrunch is a semidefinite-based solver for binary quadratic problems.  *
// *  It uses a branch-and-bound method featuring an improved semidefinite     *
// *  bounding procedure, mixed with a polyhedral approach. BiqCrunch uses     *
// *  particular input files format to describe the combinatorial problems.    *
// *                                                                           *
// *   Copyright (C) 2010-2014 Nathan Krislock, Jérôme Malick, Frédéric Roupin *
// *                                                                           *
// *                 http://www-lipn.univ-paris13.fr/BiqCrunch/                *
// *									       *
// *****************************************************************************
//									       *
//    This program is free software: you can redistribute it and/or modify     *
//    it under the terms of the GNU General Public License as published by     *
//    the Free Software Foundation, either version 3 of the License, or        *
//    (at your option) any later version.                                      *
//                                                                             *
//    This program is distributed in the hope that it will be useful,          *
//    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
//    GNU General Public License for more details.                             *
//                                                                             *
//    You should have received a copy of the GNU General Public License        *
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
//                                                                             *
// *****************************************************************************

#include <stdio.h>

#include "bb.h"
#include "biqcrunch.h"

/********************************************************/
/************ List of all global variables **************/
/********************************************************/

Problem *SP; // Original problem instance
Problem *PP; // Subproblem instance

double *SubMat; // Construction matrices (size BobSizePb)
double *SubMatLin; // Construction vector (size BobSizePb)

//=>>>>>>>> F.R. 2014
Sparse SubMatS; //Construction matrices, sparse representation

double *tempMatrix; // Construction vector (size BobSizePb + 1)

double *PX; // projected matrix computed by projSDP
double *X; // Matrix to work on : modified by proj
double *Xnew; // Stores current X
double *Xold; // Stores previous X
double *Xtest; // Stores inequality testing X

Inequality *Cuts; // vector (MaxNineqAdded) of triangle inequality constraints
Inequality *List; // list of inequalities

double *g, f; // gradient and function values
double *y; // dual variables

int nmax; // nmax is the dimension of the largest problem to be solved

double *binf; // lower bounds on the variables y
double *bsup; // upper bounds on the variables y
int *nbd; // indicates which variables are bounded

int mmax; // mmax is the maximum number of limited memory corrections.
double *wa;  // Double workspace for L-BFGS-B
int *iwa; // Integer workspace for L-BFGS-B

double *W; // contains the eigenvalues
double *Z; // contains the eigenvectors
int *ISUPPZ; // dim = 2*max(1,M), support of the eigvecs. in Z
double *WORK; // dim = LWORK
int *IWORK; // dim = LIWORK
int sizeWORK, sizeIWORK;

BiqCrunchParameters params;

int evalcount; // Number of node evaluations
double TEMPI; // initial time

// True if the algorithm stopped at root node or after a time limit
int stopped = 0;

/*
 * Best evaluation value.
 * This value keeps the value of the more promising node if the algorithm is 
 * interrupted at root node or after a time limit
 */
double best_eval;

int current_bound = BIG_NUMBER;

double scaleIneq; // scales the inequality constraints (and dual vars)
double scaleEq; // scales the equality constraints (and dual vars)

double gradInorm; // norm of the gradient of f for equality constraints
double gradEnorm; // norm of the gradient of f for equality constraints
int count; // major iteration counter
int nbit; // number of iterations in BFGS

double gradInorm; // norm of the gradient of f for equality constraints
double gradEnorm; // norm of the gradient of f for equality constraints
int count; // major iteration counter
int nbit; // number of iterations in BFGS

int count_sparse_eq = 0;
int count_sparse_ineq = 0;

double minAllIneq; // value of the minimum inequality : feasible when minAllIneq >= 0

int NIneq; // current number of inequalities used in optimization
int NumSubtracted;
int NumAdded;

FILE *output;


void allocProblem(Problem * prob, int N, int mA, int mB, int max_problem) {
	int vec_size;
	prob->n = N;
	prob->mA = mA;
	prob->mB = mB;
        prob->max_problem = max_problem;

	vec_size = (prob->n * prob->n) / 2 + prob->n;

	// Q : Objective matrix
	alloc_matrix(prob->Q, prob->n, double);

	alloc_vector(prob->Qs.i, vec_size, int);
	alloc_vector(prob->Qs.j, vec_size, int);
	alloc_vector(prob->Qs.val, vec_size, double);

	// A : Inequality constraints
	//right-hand-side vector of inequality constraints
	alloc_vector(prob->As, prob->mA, Sparse);
	// list of inequality coefficient matrices
	alloc_vector(prob->a, prob->mA, double);

	// B : Equality constraints
	//right-hand-side vector of equality constraints
	alloc_vector(prob->Bs, prob->mB, Sparse);
	// list of inequality coefficient matrices
	alloc_vector(prob->b, prob->mB, double);
}


void allocCopyProblem(Problem *d, Problem *o) {
	int s;
	d->n = o->n;
	d->mA = o->mA;
	d->mB = o->mB;
        d->max_problem = o->max_problem;

	alloc_vector(d->As, d->mA, Sparse);
	alloc_vector(d->a, d->mA, double);

	alloc_vector(d->Bs, d->mB, Sparse);
	alloc_vector(d->b, d->mB, double);

	alloc_matrix(d->Q, d->n, double);

	d->Qs.nnz = o->Qs.nnz;

	alloc_vector(d->Qs.i, d->Qs.nnz, int);
	alloc_vector(d->Qs.j, d->Qs.nnz, int);
	alloc_vector(d->Qs.val, d->Qs.nnz, double);

	for (s = 0; s < d->mB; s++) {

		d->Bs[s].nnz = o->Bs[s].nnz;
		alloc_vector(d->Bs[s].i, d->Bs[s].nnz, int);
		alloc_vector(d->Bs[s].j, d->Bs[s].nnz, int);
		alloc_vector(d->Bs[s].val, d->Bs[s].nnz, double);
	}

	for (s = 0; s < d->mA; s++) {

		d->As[s].nnz = o->As[s].nnz;
		alloc_vector(d->As[s].i, d->As[s].nnz, int);
		alloc_vector(d->As[s].j, d->As[s].nnz, int);
		alloc_vector(d->As[s].val, d->As[s].nnz, double);

	}
}



void allocSDPbound(Problem * SP) 
{
	int N = SP->n;
    int wa_length;
    int iwa_length;

	nmax = SP->mB + SP->mA + MaxNineqAdded; 
        // number of equality + inequality constraints
	mmax = 10;
	f = 0.0;

	alloc_matrix(Xnew, N, double);
	alloc_matrix(X, N, double);
	alloc_matrix(Xold, N, double);
	alloc_matrix(Xtest, N, double);

	alloc_vector(Cuts, MaxNineqAdded, Inequality);
	alloc_vector(List, params.cuts, Inequality);
	alloc_vector(g, nmax, double);
	alloc_vector(y, nmax, double);
	alloc_vector(binf, nmax, double);
	alloc_vector(bsup, nmax, double);
	alloc_vector(nbd, nmax, int);

    wa_length = (2 * mmax + 5) * nmax + 11 * mmax * mmax + 8 * mmax;
    iwa_length = 3 * nmax;

	alloc_vector(wa, wa_length, double);
	alloc_vector(iwa, iwa_length, int);

	allocProj(SP);

	alloc_matrix(PX, N, double);
}

void allocProj(Problem * P0) {
	int N = P0->n;

	sizeWORK = 26 * N;
	sizeIWORK = 10 * N;
	alloc_vector(WORK, sizeWORK, double);
	alloc_vector(IWORK, sizeIWORK, int);
	alloc_vector(W, N, double);
	alloc_matrix(Z, N, double);
	alloc_vector(ISUPPZ, 2 * N, int);

}



/*
 * Free the global memory allocated and the memory allocated to compute 
 * the bound of the nodes of the branch-and-bound tree
 */
void freeMemory() {

    // free resources of the heuristic
	BC_freeHeuristic();

    // free problem instance
	freeProblem(SP);
	free(SP);
	freeProblem(PP);
	free(PP);

    // free resources used to compute the bound
	freeSDPbound();

    // free resources of BiqCrunch
	free(tempMatrix);
	free(BobSol);
	free(SubMat);
	free(SubMatLin);
}



void freeProblem(Problem *prob) {

	int i;
	free(prob->Q);
	free(prob->Qs.i);
	free(prob->Qs.j);
	free(prob->Qs.val);

	for (i = 0; i < prob->mA; i++) {

		free(prob->As[i].i);
		free(prob->As[i].j);
		free(prob->As[i].val);
	}

	free(prob->As);
	free(prob->a);

	for (i = 0; i < prob->mB; i++) {

		free(prob->Bs[i].i);
		free(prob->Bs[i].j);
		free(prob->Bs[i].val);
	}

	free(prob->b);
	free(prob->Bs);
}


void freeSDPbound() 
{
	freeProj();
	free(Cuts);
	free(List);
	free(X);
	free(PX);
	free(Xold);
	free(Xtest);
	free(Xnew);
	free(g);
	free(y);
	free(binf);
	free(bsup);
	free(nbd);
	free(wa);
	free(iwa);
}


void freeProj() {
	free(W);
	free(Z);
	free(ISUPPZ);
	free(WORK);
	free(IWORK);

}
