// *****************************************************************************
// *                                                                           *
// *  BiqCrunch is a semidefinite-based solver for binary quadratic problems.  *
// *  It uses a branch-and-bound method featuring an improved semidefinite     *
// *  bounding procedure, mixed with a polyhedral approach. BiqCrunch uses     *
// *  particular input files format to describe the combinatorial problems.    *
// *                                                                           *
// *   Copyright (C) 2010-2014 Nathan Krislock, Jérôme Malick, Frédéric Roupin *
// *                                                                           *
// *                 http://www-lipn.univ-paris13.fr/BiqCrunch/                *
// *									       *
// *****************************************************************************
//									       *
//    This program is free software: you can redistribute it and/or modify     *
//    it under the terms of the GNU General Public License as published by     *
//    the Free Software Foundation, either version 3 of the License, or        *
//    (at your option) any later version.                                      *
//                                                                             *
//    This program is distributed in the hope that it will be useful,          *
//    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
//    GNU General Public License for more details.                             *
//                                                                             *
//    You should have received a copy of the GNU General Public License        *
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
//                                                                             *
// *****************************************************************************

#include <stdio.h>
#include <math.h>
#include <string.h>

#include "bb.h"
#include "biqcrunch.h"

extern int evalcount;
extern FILE *output;
extern BiqCrunchParameters params;
extern double TEMPI; // initial time
extern double minAllIneq; // min ineq: feasible when minAllIneq >= 0
extern Inequality *Cuts; // vector of triangle inequality constraints
extern int NIneq; // current number of inequalities used in optimization
extern int NumSubtracted;
extern int NumAdded;
extern double *PX; // projected matrix computed by projSDP
extern double scaleIneq; // scales the inequality constraints (and dual vars)
extern double scaleEq; // scales the equality constraints (and dual vars)
extern double *g, f; // gradient and function values
extern double *y; // dual variables
extern double *Xnew; // Stores current X
extern double *binf; // lower bounds on the variables y
extern double *bsup; // upper bounds on the variables y
extern int *nbd; // indicates which variables are bounded
extern int mmax; // mmax is the maximum number of limited memory corrections.
extern double *wa;  // Double workspace for L-BFGS-B
extern int *iwa; // Integer workspace for L-BFGS-B
extern double gradInorm; // norm of the gradient of f for equality constraints
extern double gradEnorm; // norm of the gradient of f for equality constraints
extern int count; // major iteration counter
extern int nbit; // number of iterations in BFGS


/********************************************************************/
/************************ SDPbound **********************************/
/********************************************************************/
double SDPbound(Problem *SP, Problem *P, double beta, double fixedvalue,
		BobNode * node, int *to_update) 
{
	int i;
	int BFGSsuccess;
	double bound;
	double oldf, newf, fdiff;

	int N = P->n;  // N is the size of the SDP matrix
	double *Matrix = P->Q;

    double ynorm;

    fprintf(output, "\n\n");
	fprintf(output,
			"************************************************************************************************\n");
	fprintf(output, 
            "                              Node %d\n", evalcount + 1);
	fprintf(output,
			"************************************************************************************************\n");
	fprintf(output, 
            "Problem size %d\n", P->n - 1);
	fflush(NULL);
	if (N == 0) {
		bound = 0.0;
		fprintf(output, "N = 0 : returning bound = %.1f\n", bound);
		return bound;
	}

	if (N == 1) {
		bound = Matrix[0];
		fprintf(output,
				"N = 1 : returning bound = Matrix[0] = %.1f\n",
				Matrix[0]);
		return bound;
	}

    NIneq = 0; // start with no cuts
    NumAdded = 0;
    NumSubtracted = 0;

	// Parameters
	int MinNiter = params.minNiter;

	// Set inital values for dynamically adjusted parameters
	double alpha = params.alpha0;
	double tol = params.tol0;

	// Compute scaling factors
	scaleIneq = 1.0 / (1.0 + sqrt(3.0));
	scaleEq = 1.0;

	// initial dual vector y = 0
	for (i = 0; i < P->mB + P->mA; i++) {
		y[i] = 0.0;
	}

	// Evaluate the function
	sim(P, NIneq, y, &f, g, alpha);

	count = 0;
	int done = 0;
	int stop_SDP_bound = 0;
	int giveup = 0;
	int prune = 0;

	if (evalcount == 0) {
		MinNiter = 20;  // Root node MinNiter is at least 20
	}

	// Main loop
	while (!done) {

        int effect = count % 4;
        switch(effect) {
            case 0 : printf("/\b"); break;
            case 1 : printf("-\b"); break;
            case 2 : printf("\\\b"); break;
            case 3 : printf("|\b"); break;
            default : break;
        }	
        
		// Update iteration counter
		count++;

		oldf = f;

		// CALL BFGS Solver
		BFGSsuccess = calllbfgsb(y, P, alpha, tol, beta, fixedvalue);

        // Compute the norm of the dual variables y
        ynorm = 0.0;
        for (i = 0; i < P->mB + P->mA + NIneq; i++) {
            ynorm += y[i]*y[i];
        }
        ynorm = sqrt(ynorm);

		if (SP->max_problem)
		 prune = ((long int) (f - fixedvalue) <= (long int) (beta - fixedvalue));
		else prune = (-ceil(-f + fixedvalue) <= (long int) (beta - fixedvalue));

		////////////////////////////////////////////////////////////////////////
		//  Heuristic 2
		////////////////////////////////////////////////////////////////////////

		double heur_val;

		if (SP->max_problem) heur_val = -BIG_NUMBER; else heur_val = BIG_NUMBER;
		
		int x[BobPbSize];  // BobPbSize = SP->n - 1;
		for (i = 0; i < BobPbSize; i++)
			x[i] = 0;

		if (params.heur_2 && !prune)
			heur_val = BC_runHeuristic(SP, P, node, x, SDP_BOUND_HEUR);

		if (!SP->max_problem) heur_val = -heur_val;

		// check if beta has been improved
		if (
			((beta - fixedvalue < heur_val) && (beta - fixedvalue == -BIG_NUMBER) && (!SP->max_problem))
			||
			((-beta + fixedvalue > heur_val) && (heur_val > -BIG_NUMBER) && (!SP->max_problem))
			||
			((beta - fixedvalue < heur_val) && (SP->max_problem))
			)
		  {
			if (BobSTAT1) {
				fprintf(output,
						"heuristic 2 Beta updated:  %lf -> %lf\n",
						-beta + fixedvalue, heur_val);
			}

			// update beta
			if (!SP->max_problem) 
                beta = -heur_val + fixedvalue;
			else 
                beta = heur_val + fixedvalue;

			// update sol
			for (i = 0; i < BobPbSize; i++) {
				node->sol.X[i] = x[i];
			}

			*to_update = 1;
		}

		///////////////////////////////////////////////////////////////////////////

		if (SP->max_problem)
            prune = ((long int) (f - fixedvalue) <= (long int) (beta - fixedvalue));
		else 
            prune = (-ceil(-f + fixedvalue) <= (long int) (beta - fixedvalue));

		newf = f;

		fdiff = newf - oldf;

		if (count >= MinNiter && !params.root) { 
			giveup = (fabs(fdiff) < 1.0) && (alpha < 1e-4)
					&& (fabs(f - beta) > 2.0);
		}

		// Add violated inequalities, remove satisfied inequalities
		if (params.withCuts) {
			if (!prune && !giveup)
				AddSubInequalities(P, y, N, params.gapCuts);
			else {
				NumAdded = 0;
				NumSubtracted = 0;
			}
		}

		stop_SDP_bound = (alpha == params.minAlpha) && (NumAdded == 0);

		if (BobSTAT1) {
			if (count == 1) {
				fprintf(output,
						"================================================================================================\n");
				fprintf(output,
						"Iter  Time    gap     alpha     tol   nbit   Enorm    Inorm    ynorm    minCuts NCuts  NSub NAdd\n");
				fprintf(output,
						"================================================================================================\n");
			}
			fprintf(output,
					"%3d %6.2f %8.3f %8.1e %8.1e %4d %8.1e %8.1e %8.1e %9.1e %5d  -%-3d +%-3d\n",
					count, temps_CPU() - TEMPI, f - beta, alpha, tol, nbit,
					gradEnorm, gradInorm, ynorm, minAllIneq, NIneq, NumSubtracted,
					NumAdded);
		}

		fflush(NULL);

		if (giveup && !prune)
			fprintf(output, "Give up ! BFGS iterations = %d Bound = %f\n",
					count, f - fixedvalue);
		if (prune)
			fprintf(output, "Prune ! BFGS iterations = %d\n", count);
		if (!BFGSsuccess)
			fprintf(output, "BFGS stopped ! iterations = %d\n", count);
		if (nbit >= params.nitermax)
			fprintf(output,
					"Maximum Iterations reached ! BFGS iterations = %d\n",
					count);

		// Test stopping criteria for adding inequalities
		done = prune || // can prune the B&B tree
				(count >= params.maxNiter) || giveup || // see giveup tests above
				stop_SDP_bound || // see stop_SDP_bound tests above
				!BFGSsuccess || // BFGS failed
				(nbit >= params.nitermax); // reached max number of BFGS iters allowed

		// Decrease alpha and tol when # of inequals added is less than params.minCuts
		if (!done && (NumAdded < params.minCuts) && (count > 1)) {
			alpha *= params.scaleAlpha;
			alpha = (alpha < params.minAlpha) ? params.minAlpha : alpha;
			tol *= params.scaleTol;
			tol = (tol < params.minTol) ? params.minTol : tol;
		}

        // Store the fractional solution in the node
        int index = 0;
        for (i = 0; i < BobPbSize; i++) {
            if (node->xfixed[i])
                node->fracsol[i] = (double) node->sol.X[i];
            else {
                // convert from -1..1 to 0..1
                node->fracsol[i] = 0.5 * (Xnew[index + (P->n - 1) * P->n] + 1.0); 
                index++;
            }
        }

	} // end main while loop

	fprintf(output,
			"================================================================================================\n");
	if (SP->max_problem) 
        fprintf(output, "Bound = %.2f, ",   f - fixedvalue); 
	else 
        fprintf(output, "Bound = %.2f, ", - f + fixedvalue); 
        
    fprintf(output, "BFGS = %d, alpha = %7.1e, tol = %7.1e, cuts = %4d, time = %.1f\n",
			count, alpha, tol, NIneq, temps_CPU() - TEMPI);

	bound = f;

	return bound;
}





/************************ calllbfgs *********************************/
int calllbfgsb(double *y, Problem *P, double alpha, double tol, double beta, 
        double fixedvalue)
{
	int i, j;
	char task[60], csave[60];
    int lsave[4], isave[44];
    double dsave[29];
	int status = 1;
	int N = P->n;
	int iprint = -1;
    int ineq;
    int mem = mmax;

    // Update count of added cuts
    NIneq = NIneq - NumSubtracted + NumAdded;

	int nvars = P->mB + P->mA + NIneq; // number of variables in BFGS

    // Initialize y
    for (ineq = 0; ineq < NIneq; ineq++) {
        y[P->mB + P->mA + ineq] = Cuts[ineq].y;
    }

	// evaluate f,g with updated y, alpha, and Inequalities (stored in vector Cuts)
	sim(P, NIneq, y, &f, g, alpha);

	// compute the bounds for y

	// equalities
	for (i = 0; i < P->mB; i++) {
		nbd[i] = 0; // y[i] unbounded
	}

	//  inequalities
	for (i = P->mB; i < nvars; i++) {
		nbd[i] = 1; // y[i] bounded from below
		binf[i] = 0.0; // lower bound is zero
	}

	//BFGS main loop
	strcpy(task, "START"); for(i=5; i<60; i++) task[i] = ' ';
	int StopBFGS = 0;
	nbit = 0; // counts number of function/gradient evaluates
	while (!StopBFGS) {
		double factr = 0.0; // CONVERGENCE: REL_REDUCTION_OF_F <= FACTR*EPSMCH
		double pgtol = 0.0; // NORM OF PROJECTED GRADIENT <= PGTOL

        /* This calls the main L-BFGS-B function */
		setulb_(&nvars, &mem, y, binf, bsup, nbd, &f, g,
				&factr, &pgtol, wa, iwa, task, &iprint,
                csave, lsave, isave, dsave);

		if (strncmp(task, "FG", 2) == 0) { // L-BFGS-B requesting new f and g
			sim(P, NIneq, y, &f, g, alpha);
			nbit++;
		} else if (strncmp(task, "NEW_X", 5) == 0) { // L-BFGS-B found new x
			// test if we should stop
			// (e.g. if bound is worse than best current solution)
			// Compute Infinity-norm of gradE = g[0:(P->mB)-1]
			gradEnorm = 0.0;
			for (i = 0; i < P->mB; i++)
				gradEnorm = (gradEnorm < fabs(g[i])) ? fabs(g[i]) : gradEnorm;
			gradEnorm /= scaleEq; // remove scaling factor when computing norm

			// Compute Infinity-norm of gradI = min( g[(P->mB):nvars-1], 0.0 )
			gradInorm = 0.0;
			for (i = P->mB; i < nvars; i++) {
				double mintemp = (g[i] > 0.0) ? 0.0 : g[i];
				gradInorm =
						(gradInorm < fabs(mintemp)) ? fabs(mintemp) : gradInorm;
			}
			gradInorm /= scaleIneq; // remove scaling factor when computing norm

			if (((long int) (f - fixedvalue) <= (long int) (beta - fixedvalue))
					|| // can prune the B&B tree
					(gradEnorm < tol && gradInorm < tol)
					|| (nbit > params.nitermax)) {
				strcpy(task, "STOP"); for(i=4; i<60; i++) task[i] = ' ';
			}
		} else { // L-BFGS-B has terminated
			StopBFGS = 1;
			if (strncmp(task, "STOP", 4) != 0) {
				task[4] = '\0';
				fprintf(output, "task = %s\n", task);
				status = 0;
			}
		}
	}
	// Compute Xnew from PX :  Xnew = PX/alpha
	double alphainv = 1.0 / alpha;
	for (i = 0; i < N; i++) {
		for (j = 0; j < i; j++) {
			Xnew[i + j * N] = PX[i + j * N] * alphainv;
			Xnew[j + i * N] = Xnew[i + j * N];
		}
		Xnew[i + i * N] = PX[i + i * N] * alphainv;
	}

	return status;
}
