// *****************************************************************************
// *                                                                           *
// *  BiqCrunch is a semidefinite-based solver for binary quadratic problems.  *
// *  It uses a branch-and-bound method featuring an improved semidefinite     *
// *  bounding procedure, mixed with a polyhedral approach. BiqCrunch uses     *
// *  particular input files format to describe the combinatorial problems.    *
// *                                                                           *
// *   Copyright (C) 2010-2014 Nathan Krislock, Jérôme Malick, Frédéric Roupin *
// *                                                                           *
// *                 http://www-lipn.univ-paris13.fr/BiqCrunch/                *
// *									       *
// *****************************************************************************
//									       *
//    This program is free software: you can redistribute it and/or modify     *
//    it under the terms of the GNU General Public License as published by     *
//    the Free Software Foundation, either version 3 of the License, or        *
//    (at your option) any later version.                                      *
//                                                                             *
//    This program is distributed in the hope that it will be useful,          *
//    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
//    GNU General Public License for more details.                             *
//                                                                             *
//    You should have received a copy of the GNU General Public License        *
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
//                                                                             *
// *****************************************************************************

#include <stdio.h>
#include <math.h>

#include "bb.h"
#include "biqcrunch.h"

extern Problem *SP; // problem
extern Problem *PP; // subproblem (some variables are fixed) : PP is made from SP
extern double *SubMat; // Construction matrices (size BobSizePb)
extern double *SubMatLin; // Construction vector (size BobSizePb)

//=> F.R. 2014 Sparse
extern Sparse SubMatS; //Construction matrices, sparse representation

extern double *tempMatrix; // Construction vector (size BobSizePb + 1)
extern BiqCrunchParameters params;
extern FILE *output;
extern int evalcount;

/*
 * Evaluate a specific node.
 * This function computes the bound of a specific node
 * @param n: the node of the brach-and-bound tree
 * @return the bound of the node n
 */
double Evaluate(BobNode *node) {

	double bound;
	double Constant;
	double norm;
	int SubN;
	int N;
	int nz;
	int i, j, k, s;
    	double scale_factor;

/*
///////////////////////////////////////////////////////////////////////////
	 Compute the subproblem removing the rows and the columns of the 
         fixed variables from each matrix (Q, B[1..mB], A[1..mA])
///////////////////////////////////////////////////////////////////////////
*/

        // number of non-fixed variables in the node
	// Updated by F.R. 2014 : for some problems several variables may be fixed in the same node (e.g. QAP).
	// So now it is computed from node->xfixed directly (and not only from node->level). 

//	SubN = BobPbSize - node->level; 

	SubN = BobPbSize;
	for (i = 0; i < BobPbSize; i++)  if (node->xfixed[i]) SubN--; 

///////////////////////////////////////////////////////////////////////////////
//	Objective function of the subproblem : PP->Q, PP->Qs, made from SP->Qs
///////////////////////////////////////////////////////////////////////////////

//      We shall use dense representation in future calculus so no need to build
//      a sparse representation (getSubMatrixSparse))

    // Build the quadratic part of the matrix Q
	getSubMatrix(SubMat, &(SP->Qs), node->xfixed, BobPbSize); 

    // Build the linear part of the matrix Q
	getLinear(SubMatLin, &(SP->Qs), node->sol.X, node->xfixed, BobPbSize); 

    // Build the constant part of the matrix Q
	Constant = SP->Q[(SP->n - 1) + (SP->n - 1) * SP->n]; 

	// Add linear terms to diagonal of submatrix
	for (i = 0; i < SubN; i++)
		SubMat[i + i * SubN] += SubMatLin[i];

	// Build current objective function PP->Q
	N = SubN + 1;
	PP->n = N;
	PP->Q[(N - 1) + (N - 1) * N] = Constant; // Constant part
	for (i = 0; i < SubN; i++)
		for (j = 0; j < SubN; j++)
			if (i != j) {
				PP->Q[i + j * N] = SubMat[i + j * SubN];
			} else {
				PP->Q[i + j * N] = 0.0;
			} // Quadratic part
	for (i = 0; i < SubN; i++) {
		PP->Q[i + (N - 1) * N] = SubMat[i + i * SubN]; // Linear part
		PP->Q[(N - 1) + i * N] = SubMat[i + i * SubN];
	}

    // Compute the sparse matrix from the dense one
	sparseMatrix(&(PP->Qs), PP->Q, PP->n); 

    ///////////////////////////////////////////////////////////////////////
    //===================================
    // 2. Equality constraints B[1..mB]
    //===================================
    //////////////////////////////////////////////////////////////////////

	for (s = 0; s < SP->mB; s++) {
		PP->b[s] = SP->b[s];
		norm = 0;
//=> F.R. 2014 Sparse
		norm = getSubMatrixSparse(&SubMatS, &(SP->Bs[s]), node->xfixed, BobPbSize);
		//Build the linear part and the constant part
		Constant = getConstant(&(SP->Bs[s]), node->sol.X, node->xfixed, BobPbSize);
		norm = sqrt(norm);		
		getLinear(SubMatLin, &(SP->Bs[s]), node->sol.X, node->xfixed, BobPbSize);
//		norm = sqrt(norm2+norm+Constant*Constant);		

//=> F.R. 2014 Sparse
		if (params.scaling)
            // 1 / (1 + |b| + ||B||)
			scale_factor = 1.0 / (1.0 + (SP->b[s] > 0 ? SP->b[s] : -SP->b[s]) + norm); 
		else
			scale_factor = 1.0;

// Build the final submatrix
	nz = 0;

        // Quadratic part
	for (k = 0; k < SubMatS.nnz; k++) {
		if (fabs(SubMatS.val[k]) > 0.0) {
			PP->Bs[s].i[nz] = SubMatS.i[k];
			PP->Bs[s].j[nz] = SubMatS.j[k];
			PP->Bs[s].val[nz] = scale_factor*SubMatS.val[k];
			nz++;
		}
	}
        
	// Linear part
	for (i = 0; i < SubN; i++) {
		if (fabs(SubMatLin[i]) > 0.0) {
			PP->Bs[s].i[nz] = SubN; 
			PP->Bs[s].j[nz] = i;
			PP->Bs[s].val[nz] = scale_factor*SubMatLin[i];
			nz++;
		}
	}
	
        // new constant value on left-side
	if (fabs(Constant)>0.0) {
		PP->Bs[s].i[nz] = SubN;   
		PP->Bs[s].j[nz] = SubN;
		PP->Bs[s].val[nz] = scale_factor*Constant;
		nz++;
	}	

	// Final number of non-zero elements of the submatrix
	PP->Bs[s].nnz = nz;

// Scale the right-side of the contraint
		PP->b[s] *= scale_factor;

      } // end (for s:1..mB)


    /////////////////////////////////////////////////////////////
    //===========================
    // 3. Inequality constraints A[1..mA]
    /////////////////////////////////////////////////////////////

	for (s = 0; s < SP->mA; s++) {
		PP->a[s] = SP->a[s];
		norm = 0;
//=> F.R. 2014 Sparse
		norm = getSubMatrixSparse(&SubMatS, &(SP->As[s]), node->xfixed, BobPbSize);

		//Build the linear part and the constant part
		Constant = getConstant(&(SP->As[s]), node->sol.X, node->xfixed, BobPbSize);
		norm = sqrt(norm);		
		getLinear(SubMatLin, &(SP->As[s]), node->sol.X, node->xfixed, BobPbSize);
//		norm = sqrt(norm2+norm+Constant*Constant);		

//=> F.R. 2014 Sparse
		if (params.scaling)
            // 1 / (1 + |a| + ||A||)
			scale_factor = 1.0 / (1.0 + (SP->a[s] > 0 ? SP->a[s] : -SP->a[s]) + norm); 
		else
			scale_factor = 1.0;

// Build the final submatrix
	nz = 0;

        // Quadratic part
	for (k = 0; k < SubMatS.nnz; k++) {
		if (fabs(SubMatS.val[k]) > 0.0) {
			PP->As[s].i[nz] = SubMatS.i[k];
			PP->As[s].j[nz] = SubMatS.j[k];
			PP->As[s].val[nz] = scale_factor*SubMatS.val[k];
			nz++;
		}
	}
        
	// Linear part
	for (i = 0; i < SubN; i++) {
		if (fabs(SubMatLin[i]) > 0.0) {
			PP->As[s].i[nz] = SubN; 
			PP->As[s].j[nz] = i;
			PP->As[s].val[nz] = scale_factor*SubMatLin[i];
			nz++;
		}
	}
	
        // new constant value on left-side

	if(fabs(Constant)>0.0) {
		PP->As[s].i[nz] = SubN;   
		PP->As[s].j[nz] = SubN;
		PP->As[s].val[nz] = scale_factor*Constant;	
		nz++;
	}

	// Final number of non-zero elements of the submatrix
	PP->As[s].nnz = nz;

// Scale the right-side of the contraint
	PP->a[s] *= scale_factor;
	}

    //===========================================================================
    //===========================================================================

    // Compute SDPbeta
	int beta = Bob_ULBGet();

	double fixedvalue = 0.0;

    // linear part (last column)
	for (i = 0; i < BobPbSize; i++)
		if (node->xfixed[i])
			fixedvalue -= (double) (2 * (node->sol.X[i]) - 1.0)
					* (SP->Q[i + (SP->n - 1) * (SP->n)]
							+ SP->Q[(SP->n - 1) + i * (SP->n)]);

    // quadratic part
	for (i = 0; i < BobPbSize; i++) {
		for (j = 0; j < BobPbSize; j++) {
			if ((node->xfixed[i]) && (node->xfixed[j]))
				fixedvalue -= (double) (2 * (node->sol.X[i]) - 1.0)
						* (double) (2 * (node->sol.X[j]) - 1.0)
						* SP->Q[i + j * SP->n];
		}
	}

	double SDPbeta = -(double) beta + fixedvalue;

	int to_update = 0;

    //===============  SDPbound  ==================//
	bound = SDPbound(SP, PP, SDPbeta, fixedvalue, node, &to_update);
	bound -= fixedvalue;
	bound = -bound; // since we are solving a "min" problem
	evalcount++;
    //===============  SDPbound  ==================//

	if (to_update) {
		double heur_val = BC_evaluateSolution(node->sol.X);
		addSolution(node, -heur_val, node->sol.X, NEWSOL_HEUR2);
	}

    // Check if we can prune
	int intbound;

   // Checking int overflow : useful to test infeasibility
	if( bound > BIG_NUMBER) bound = BIG_NUMBER;
	if (bound < -BIG_NUMBER) bound = -BIG_NUMBER;

   //	Rounding the bound
	if (SP->max_problem) intbound = (int) bound; else intbound = ceil(bound);

	if (!Bob_ULBSup(intbound)) { // (intbound >= Bob_ULBGet())
		// BobULB is not superior to intbound
		fprintf(output, "Prune !\n");
	} else {

		////////////////////////////////////////////////////////////////////////
		//  Heuristic 3
		////////////////////////////////////////////////////////////////////////

		double heur_val = 0.0;
		int temp_x[BobPbSize];
		for (i = 0; i < BobPbSize; i++)
			if (node->xfixed[i])
				temp_x[i] = node->sol.X[i];
			else
				temp_x[i] = 0;
		if (params.heur_3)
			heur_val = BC_runHeuristic(SP, PP, node, temp_x, ROUNDING_HEUR);

		addSolution(node, -heur_val, temp_x, NEWSOL_HEUR3);

		////////////////////////////////////////////////////////////////////////
		// End of Heuristic 3
		////////////////////////////////////////////////////////////////////////

		if (!Bob_ULBSup(intbound)) { // (intbound >= Bob_ULBGet())
			// BobULB is not superior to intbound
			fprintf(output, "Prune !\n");
		}
	}

	return bound;
}

//=> F.R. 2014 Sparse
double getSubMatrixSparse(Sparse *subM, Sparse *sparseMatrix, int *fixed, int N) {
	double sum = 0.0;
	int i,k;
	int Nfixed = 0;
//	int Nfree;
	int offset[N];
	int count;

	count = 0;
	for (i = 0; i < N; i++) {
		if (fixed[i]) {
			offset[i] = 0;
			count++;
			Nfixed++;
		} else
			offset[i] = count;
	}

	k = 0; //current number of non-zero elements of the submatrix SubM

	for (i = 0; i < sparseMatrix->nnz; i++)
		if (sparseMatrix->i[i] < N && sparseMatrix->j[i] < N) {
			if (!fixed[sparseMatrix->i[i]] && !fixed[sparseMatrix->j[i]]) {
				int ii = sparseMatrix->i[i] - offset[sparseMatrix->i[i]];
				int jj = sparseMatrix->j[i] - offset[sparseMatrix->j[i]];
				double val = sparseMatrix->val[i];
				if (jj<=ii) {
					subM->i[k] = ii;  //[ii + jj * Nfree] = val;
					subM->j[k] = jj; //[jj + ii * Nfree] = val;
				}
				else {
					subM->i[k] = jj;  //[ii + jj * Nfree] = val;
					subM->j[k] = ii; //[jj + ii * Nfree] = val;
				} 
				subM->val[k] = val;
				k++; // Increase the number of non-zero elements of SubM
				sum += (val * val);
				if (ii != jj) sum += (val * val); // off-diagonal : counted twice !
			}
		}

	subM->nnz = k; //final number of non-zero elements of the submatrix

	return sum; // return the squared norm of the matrix
}

double getSubMatrix(double * subM, Sparse * sparseMatrix, int * fixed, int N) {
	double sum = 0.0;
	int i;
	int Nfixed = 0;
	int Nfree;
	int offset[N];
	int count;

	count = 0;
	for (i = 0; i < N; i++) {
		if (fixed[i]) {
			offset[i] = 0;
			count++;
			Nfixed++;
		} else
			offset[i] = count;
	}

	Nfree = N - Nfixed;

	// initialize the submatrix
	for (i = 0; i < Nfree * Nfree; i++) {
		subM[i] = 0.0;
	}

	for (i = 0; i < sparseMatrix->nnz; i++)
		if (sparseMatrix->i[i] < N && sparseMatrix->j[i] < N) {
			if (!fixed[sparseMatrix->i[i]] && !fixed[sparseMatrix->j[i]]) {
				int ii = sparseMatrix->i[i] - offset[sparseMatrix->i[i]];
				int jj = sparseMatrix->j[i] - offset[sparseMatrix->j[i]];
				double val = sparseMatrix->val[i];
				subM[ii + jj * Nfree] = val;
				subM[jj + ii * Nfree] = val;
				sum += (val * val);
				if (ii != jj)
					sum += (val * val);
			}
		}

	return sum;
}


double getConstant(Sparse * sparseMatrix, int * sol, int * fixed, int N) {

	double C = 0.0;
	int i;

	for (i = 0; i < sparseMatrix->nnz; i++) {
		int ii = sparseMatrix->i[i];
		int jj = sparseMatrix->j[i];
		double val = sparseMatrix->val[i];

		if (ii < N && jj < N) {
			if (fixed[ii] && fixed[jj]) {
				double tmp = (double) (2 * sol[ii] - 1.0)
						* (double) (2 * sol[jj] - 1.0) * val;
				C += tmp;
				if (ii != jj)
					C += tmp;
			}
		} else if (ii == N && jj == N) {
			C += val;
		} else if (jj == N) {
			if (fixed[ii]) {
				C += (double)(2 * sol[ii] - 1.0) * (val + val);
			}
		} else if (ii == N) {
			if (fixed[jj]) {
				C += (double)(2 * sol[jj] - 1.0) * (val + val);
			}
		}
	}
	return C;
}



double getLinear(double * subMatrixLin, Sparse * sparseMatrix, int * sol,
		int * fixed, int N) {
	double sum = 0;
	int i;

	int Nfixed = 0;
	int Nfree;
	int offset[N];
	int count;

	count = 0;
	for (i = 0; i < N; i++) {
		if (fixed[i]) {
			offset[i] = 0;
			count++;
			Nfixed++;
		} else
			offset[i] = count;
	}

	Nfree = N - Nfixed;

	// initialize the submatrix
	for (i = 0; i < Nfree; i++) {
		subMatrixLin[i] = 0.0;
	}
	for (i = 0; i < sparseMatrix->nnz; i++) {
		int ii = sparseMatrix->i[i];
		int jj = sparseMatrix->j[i];
		double val = sparseMatrix->val[i];

		if (ii < N && !fixed[ii] && jj == N) {
			subMatrixLin[ii - offset[ii]] += val;
//=> F.R. 2014 Sparse
			sum += val*val;	
		} else if (jj < N && !fixed[jj] && ii == N) {
			subMatrixLin[jj - offset[jj]] += val;
//=> F.R. 2014 Sparse
			sum += val*val;
		} else if (jj < N && ii < N) {
			if (!fixed[ii] && fixed[jj]) {
				subMatrixLin[ii - offset[ii]] += (double) (2 * sol[jj] - 1.0)
						* val;
//=> F.R. 2014 Sparse
			sum += (double) (2 * sol[jj] - 1.0)*val*(2 * sol[jj] - 1.0)*val;
			} else if (!fixed[jj] && fixed[ii]) {
				subMatrixLin[jj - offset[jj]] += (double) (2 * sol[ii] - 1.0)
						* val;
//=> F.R. 2014 Sparse
			sum += (double) (2 * sol[ii] - 1.0)*val*(2 * sol[ii] - 1.0)*val;
			}

		}

	}

	return 2*sum; // Linear terms will be off-diagonal => *2
}

